import { statSync } from 'fs';
import { join, relative, dirname, basename, extname, resolve } from 'upath';
import { defineNuxtModule, resolveAlias, addWebpackPlugin, addVitePlugin } from '@nuxt/kit';
import globby from 'globby';
import { splitByCase, pascalCase } from 'scule';
import { createUnplugin } from 'unplugin';
import { parseURL, parseQuery } from 'ufo';

function sortDirsByPathLength({ path: pathA }, { path: pathB }) {
  return pathB.split(/[\\/]/).filter(Boolean).length - pathA.split(/[\\/]/).filter(Boolean).length;
}
function hyphenate(str) {
  return str.replace(/\B([A-Z])/g, "-$1").toLowerCase();
}
async function scanComponents(dirs, srcDir) {
  const components = [];
  const filePaths = new Set();
  const scannedPaths = [];
  for (const { path, pattern, ignore = [], prefix, extendComponent, pathPrefix, level, prefetch = false, preload = false } of dirs.sort(sortDirsByPathLength)) {
    const resolvedNames = new Map();
    for (const _file of await globby(pattern, { cwd: path, ignore })) {
      const filePath = join(path, _file);
      if (scannedPaths.find((d) => filePath.startsWith(d))) {
        continue;
      }
      if (filePaths.has(filePath)) {
        continue;
      }
      filePaths.add(filePath);
      const prefixParts = [].concat(prefix ? splitByCase(prefix) : [], pathPrefix !== false ? splitByCase(relative(path, dirname(filePath))) : []);
      let fileName = basename(filePath, extname(filePath));
      if (fileName.toLowerCase() === "index") {
        fileName = pathPrefix === false ? basename(dirname(filePath)) : "";
      }
      const fileNameParts = splitByCase(fileName);
      const componentNameParts = [];
      while (prefixParts.length && (prefixParts[0] || "").toLowerCase() !== (fileNameParts[0] || "").toLowerCase()) {
        componentNameParts.push(prefixParts.shift());
      }
      const componentName = pascalCase(componentNameParts) + pascalCase(fileNameParts);
      if (resolvedNames.has(componentName)) {
        console.warn(`Two component files resolving to the same name \`${componentName}\`:

 - ${filePath}
 - ${resolvedNames.get(componentName)}`);
        continue;
      }
      resolvedNames.set(componentName, filePath);
      const pascalName = pascalCase(componentName).replace(/["']/g, "");
      const kebabName = hyphenate(componentName);
      const shortPath = relative(srcDir, filePath);
      const chunkName = "components/" + kebabName;
      let component = {
        filePath,
        pascalName,
        kebabName,
        chunkName,
        shortPath,
        export: "default",
        global: Boolean(global),
        level: Number(level),
        prefetch: Boolean(prefetch),
        preload: Boolean(preload)
      };
      if (typeof extendComponent === "function") {
        component = await extendComponent(component) || component;
      }
      const definedComponent = components.find((c) => c.pascalName === component.pascalName);
      if (definedComponent && component.level < definedComponent.level) {
        Object.assign(definedComponent, component);
      } else if (!definedComponent) {
        components.push(component);
      }
    }
    scannedPaths.push(path);
  }
  return components;
}

const loaderPlugin = createUnplugin((options) => ({
  name: "nuxt-components-loader",
  enforce: "post",
  transformInclude(id) {
    const { pathname, search } = parseURL(id);
    const query = parseQuery(search);
    return pathname.endsWith(".vue") && (query.type === "template" || !search);
  },
  transform(code) {
    return transform(code, options.getComponents());
  }
}));
function findComponent(components, name) {
  return components.find(({ pascalName, kebabName }) => [pascalName, kebabName].includes(name));
}
function transform(content, components) {
  let num = 0;
  let imports = "";
  const map = new Map();
  const newContent = content.replace(/ _resolveComponent\("(.*?)"\)/g, (full, name) => {
    const component = findComponent(components, name);
    if (component) {
      const identifier = map.get(component) || `__nuxt_component_${num++}`;
      map.set(component, identifier);
      imports += `import ${identifier} from "${component.filePath}";`;
      return ` ${identifier}`;
    }
    return full;
  });
  return `${imports}
${newContent}`;
}

const isPureObjectOrString = (val) => !Array.isArray(val) && typeof val === "object" || typeof val === "string";
const isDirectory = (p) => {
  try {
    return statSync(p).isDirectory();
  } catch (_e) {
    return false;
  }
};
const module = defineNuxtModule({
  name: "components",
  defaults: {
    dirs: ["~/components"]
  },
  setup(options, nuxt) {
    let componentDirs = [];
    let components = [];
    nuxt.hook("app:resolve", async () => {
      await nuxt.callHook("components:dirs", options.dirs);
      componentDirs = options.dirs.filter(isPureObjectOrString).map((dir) => {
        const dirOptions = typeof dir === "object" ? dir : { path: dir };
        const dirPath = resolveAlias(dirOptions.path, nuxt.options.alias);
        const transpile = typeof dirOptions.transpile === "boolean" ? dirOptions.transpile : "auto";
        const extensions = dirOptions.extensions || ["vue"];
        dirOptions.level = Number(dirOptions.level || 0);
        const present = isDirectory(dirPath);
        if (!present && dirOptions.path !== "~/components") {
          console.warn("Components directory not found: `" + dirPath + "`");
        }
        return {
          ...dirOptions,
          enabled: true,
          path: dirPath,
          extensions,
          pattern: dirOptions.pattern || `**/*.{${extensions.join(",")},}`,
          ignore: [
            "**/*.stories.{js,ts,jsx,tsx}",
            "**/*{M,.m,-m}ixin.{js,ts,jsx,tsx}",
            "**/*.d.ts",
            ...dirOptions.ignore || []
          ],
          transpile: transpile === "auto" ? dirPath.includes("node_modules") : transpile
        };
      }).filter((d) => d.enabled);
      nuxt.options.build.transpile.push(...componentDirs.filter((dir) => dir.transpile).map((dir) => dir.path));
    });
    nuxt.hook("app:templates", async (app) => {
      components = await scanComponents(componentDirs, nuxt.options.srcDir);
      await nuxt.callHook("components:extend", components);
      if (!components.length) {
        return;
      }
      app.templates.push({
        filename: "components.mjs",
        src: resolve(__dirname, "runtime/components.tmpl.mjs"),
        options: { components }
      });
      app.templates.push({
        filename: "components.d.ts",
        write: true,
        getContents: () => `// Generated by components discovery
declare module 'vue' {
  export interface GlobalComponents {
${components.map((c) => `  '${c.pascalName}': typeof import('${relative(nuxt.options.buildDir, c.filePath)}')['${c.export}']`).join(",\n")}
  }
}

export {}`
      });
      app.plugins.push({ src: "#build/components" });
    });
    nuxt.hook("prepare:types", ({ references }) => {
      references.push({ path: resolve(nuxt.options.buildDir, "components.d.ts") });
    });
    nuxt.hook("builder:watch", async (event, path) => {
      if (!["add", "unlink"].includes(event)) {
        return;
      }
      const fPath = resolve(nuxt.options.rootDir, path);
      if (componentDirs.find((dir) => fPath.startsWith(dir.path))) {
        await nuxt.callHook("builder:generateApp");
      }
    });
    if (!nuxt.options.dev) {
      const options2 = { getComponents: () => components };
      addWebpackPlugin(loaderPlugin.webpack(options2));
      addVitePlugin(loaderPlugin.vite(options2));
    }
  }
});

export { module as default };
